#pragma once

#include <memory>
#include <vector>
#include <iostream>

#include <cpplogging/types.h>

namespace cpplogging
{

    /**
     * @brief Class to consume log entries.
     */
    struct consumer
    {
    private:
        using format_type = std::vector<std::string>;

    protected:
        std::string     _name;      //!< name of the consumer
        format_type     _format;    //!< parsed logging format

    public:
        /**
         * @brief Get the default log format.
         */
        static inline const std::string& default_format();

    protected:
        /**
         * @brief Parse the given format and generate a pre-compiled list of strings.
         *
         * @param[in] format    Format to parse.
         *
         * @return              Precompiled format.
         */
        static format_type parse_format(const std::string& format);

        /**
         * @brief Write a log entry to the passed stream using the precompiled format.
         *
         * @param[in] e     Entry to write to stream.
         * @param[in] f     Precompiled format to use for writing.
         * @param[in] os    Stream to write entry to.
         */
        static void write_formatted(const log_entry& e, const format_type& f, std::ostream& os);

    public:
        /**
         * @brief Constructor.
         *
         * @param[in] name          Name of the consumer.
         * @param[in] format        Format to use for logging.
         */
        consumer(
            const std::string& name,
            const std::string& format = default_format());

        /**
         * @breif Destructor.
         */
        virtual ~consumer();

        /**
         * @brief Get the name of the logger.
         */
        inline const std::string& name() const;

        /**
         * @brief Consume a log entry.
         *
         * @param[in] entry Log entry to consume.
         */
        virtual void write_entry(const log_entry_ptr_s& entry) const = 0;
    };

    using consumer_ptr_u = std::unique_ptr<consumer>;

}

#include "consumer.inl"
