#pragma once

#include <set>
#include <map>
#include <list>
#include <string>

#include <cpplogging/config.h>

#include "rule.h"
#include "logger_impl.h"
#include "consumer/consumer.h"

#ifdef CPPLOGGING_HAS_NLOHMANN_JSON
    #define CPPLOGGING_HAS_LOAD_CONFIG
#endif

namespace cpplogging
{

    /**
     * @brief Wrapps the static configuration methods for the logger engine.
     *
     * The logging engine has two modes: logging and config.
     *
     * In the config mode, the user can add rules, consumer and matcher.
     * In the logging mode the engine is not thread-safe and must be accessed
     * from only one thread!
     *
     * In the logging mode the user can only create new loggers or log messages.
     * The logging mode is thread-safe. The user can use as many thread as he want,
     * to log messages.
     */
    struct manager
    {
    private:
        struct consumer_wrapper
        {
            consumer&       _consumer;  //!< consumer to use for operations
            consumer_ptr_u  _storage;   //!< storage to store owner object

            /**
             * @brief Construtor to create a consumer wrapper by using the consumer reference.
             */
            inline consumer_wrapper(consumer& p_consumer);

            /**
             * @brief Construtor to create a consumer wrapper. The wrapper takes the ownership of the consumer.
             */
            inline consumer_wrapper(consumer_ptr_u&& p_consumer);

            /**
             * @brief Operator to compare the wrapper to a other wrapper
             */
            inline bool operator<(const consumer_wrapper& other) const;

            /**
             * @brief Operator to access the consumer behind the wrapper.
             */
            inline consumer* operator->();

            /**
             * @brief Operator to access the consumer behind the wrapper.
             */
            inline const consumer* operator->() const;

            /**
             * @brief Operator to access the consumer behind the wrapper.
             */
            inline consumer& operator*();

            /**
             * @brief Operator to access the consumer behind the wrapper.
             */
            inline const consumer& operator*() const;
        };

        using logger_impl_ptr_u = std::unique_ptr<logger_impl>;
        using logger_map        = std::map<std::string, logger_impl_ptr_u>;
        using rule_list         = std::list<rule>;
        using consumer_set      = std::set<consumer_wrapper>;

    private:
        static std::mutex   _mutex;     //!< mutex to protect the engine
        static logger_impl  _default;   //!< default logger
        static logger_map   _logger;    //!< normal logger instances
        static rule_list    _rules;     //!< rules to combine loggers with consumers
        static consumer_set _consumer;  //!< consumers registered

    public:
        /**
         * @brief Get the instance of the logger with the requested name.
         *
         * @param[in] name  Name of the logger to get the instance for.
         *
         * @return          Instance of the logger with the requested name.
         */
        static logger& get_logger(const std::string& name = "");

        /**
         * @brief Register a new consumer in the logging framework.
         *
         * @param[in] consumer  Reference of the consumer to register.
         */
        static void register_consumer(consumer& consumer);

        /**
         * @brief Register a new consumer in the logging framework.
         *
         * @param[in] consumer  Consumer to register (the logging framework will own the consumer).
         *
         * @return              Pointer to the consumer.
         */
        static const consumer* register_consumer(consumer_ptr_u&& consumer);

        /**
         * @brief Unregister a consumer from the logging framework.
         *
         * @param[in] consumer  Reference of the consumer to unregister
         */
        static void unregister_consumer(consumer& consumer);

        /**
         * @brief Unregister a consumer from the logging framework.
         *
         * @param[in] consumer  Consumer to unregister from the framework.
         */
        static void unregister_consumer(consumer* consumer);

        /**
         * @brief Define a new rule in the logging framework.
         *
         * @param[in] logger_matcher    Matcher to use for matching loggers.
         * @param[in] consumer_matcher  Matcher to use for matching consumers.
         * @param[in] min_level         Minimum log level.
         * @param[in] max_level         Maximum log level.
         *
         * @return                      Return the handle to the rule (for removing).
         */
        static rule_handle define_rule(
            matcher_ptr_u&& logger_matcher,
            matcher_ptr_u&& consumer_matcher,
            log_level min_level = log_level::debug,
            log_level max_level = log_level::error);

        /**
         * @brief Remove a rule from the logging framework.
         *
         * @param[in] handle    Handle of the rule to remove.
         */
        static void undefine_rule(rule_handle handle);

        /**
         * @brief Remove all consumers, rules, matchers and loggers.
         */
        static void reset();

    #ifdef CPPLOGGING_HAS_LOAD_CONFIG
        /**
         * @brief Load current configuration from file.
         *
         * @param[in] filename  Filename to load configuration from.
         */
        static void load(const std::string& filename);
    #endif

    private:
        /**
         * @brief Initialize the given logger instance.
         */
        static logger_impl& init_logger(logger_impl& logger);
    };

}
