#pragma once

#include <set>
#include <mutex>

#include "matcher/matcher.h"
#include "consumer/consumer.h"

namespace cpplogging
{

    struct rule
    {
    private:
        mutable std::mutex          _mutex;
        std::set<const consumer*>   _consumer;

    public:
        matcher_ptr_u  logger_matcher;      //!< matcher to use for logger matching
        matcher_ptr_u  consumer_matcher;    //!< matcher to use for consumer matching
        log_level      min_level;           //!< minimum log level
        log_level      max_level;           //!< maximum log level

        /**
         * @brief Constructor.
         *
         * @param[in] lm    Matcher to use for logger matching.
         * @param[in] cm    Matcher to use for consumer matching.
         * @param[in] min   Minimum log level.
         * @param[in] max   Maximum log level.
         */
        inline rule(matcher_ptr_u&& lm, matcher_ptr_u&& cm, log_level min, log_level max);

        /**
         * @brief CHeck if the passed log level is enabled in this rule.
         *
         * @param[in] level Level to check.
         *
         * @retval true     If the log level is enabled for this rule.
         * @retval false    If the log level is not enabled for this rule.
         */
        inline bool is_enabled(log_level level) const;

        /**
         * @brief Add a consumer to this rule.
         *
         * @param[in] consumer  Consumer to add to the rule.
         */
        inline void add_consumer(const consumer& consumer);

        /**
         * @brief Remove a consumer from this rule.
         *
         * @param[in] consumer  Consumer to remove from the rule.
         */
        inline void remove_consumer(const consumer& consumer);

        /**
         * @brief Write the passed log entry to all registered consumers.
         *
         * @param[in] entry Entry to write to the consumers.
         */
        inline void write_entry(const log_entry_ptr_s& entry) const;
    };

    using rule_handle = void *;

}

#include "rule.inl"
