#include <fstream>

#include <cpplogging/manager/manager.h>
#include <cpplogging/manager/manager.inl>
#include <cpplogging/manager/matcher/matcher_all.h>
#include <cpplogging/manager/matcher/matcher_regex.h>
#include <cpplogging/manager/matcher/matcher_default_logger.h>
#include <cpplogging/manager/consumer/consumer_stream.h>

#ifdef CPPLOGGING_HAS_NLOHMANN_JSON
    #include <nlohmann/json.hpp>
#endif

using namespace ::cpplogging;

std::mutex              manager::_mutex;
logger_impl             manager::_default("");
manager::logger_map     manager::_logger;
manager::rule_list      manager::_rules;
manager::consumer_set   manager::_consumer;

logger& manager::get_logger(const std::string& name)
{
    std::lock_guard<std::mutex> lg(_mutex);
    if (name.empty())
        return _default;

    auto it = _logger.find(name);
    if (it == _logger.end())
    {
        it = _logger.emplace_hint(it, name, std::make_unique<logger_impl>(name));
        manager::init_logger(*it->second);
    }

    return *it->second;
}

void manager::register_consumer(consumer& consumer)
{
    std::lock_guard<std::mutex> lg(_mutex);
    auto it = _consumer.emplace(consumer).first;
    auto& c = **it;
    for (auto& rule : _rules)
    {
        if (rule.consumer_matcher->match(c))
            rule.add_consumer(c);
    }
}

const consumer* manager::register_consumer(consumer_ptr_u&& consumer)
{
    std::lock_guard<std::mutex> lg(_mutex);
    auto it = _consumer.emplace(std::move(consumer)).first;
    auto& ret = **it;
    for (auto& rule : _rules)
    {
        if (rule.consumer_matcher->match(ret))
            rule.add_consumer(ret);
    }
    return &ret;
}

void manager::unregister_consumer(consumer& consumer)
{
    std::lock_guard<std::mutex> lg(_mutex);
    unregister_consumer(&consumer);
}

void manager::unregister_consumer(consumer* consumer)
{
    std::lock_guard<std::mutex> lg(_mutex);
    auto it = _consumer.find(consumer_wrapper(*consumer));
    if (it == _consumer.end())
        return;

    auto& c = **it;
    for (auto& rule : _rules)
        rule.remove_consumer(c);
}

rule_handle manager::define_rule(
    matcher_ptr_u&& logger_matcher,
    matcher_ptr_u&& consumer_matcher,
    log_level min_level,
    log_level max_level)
{
    std::lock_guard<std::mutex> lg(_mutex);

    /* create rule */
    _rules.emplace_back(std::move(logger_matcher), std::move(consumer_matcher), min_level, max_level);
    auto& rule = _rules.back();

    /* update consumers */
    for (auto& c : _consumer)
    {
        if (rule.consumer_matcher->match(*c))
            rule.add_consumer(*c);
    }

    /* update default logger */
    if (rule.logger_matcher->match(_default))
        _default.add_rule(rule);

    /* update loggers */
    for (auto& l : _logger)
    {
        if (rule.logger_matcher->match(*l.second))
            l.second->add_rule(rule);
    }
    return &rule;
}

void manager::undefine_rule(rule_handle handle)
{
    std::lock_guard<std::mutex> lg(_mutex);

    /* find the rule in the list */
    auto r = static_cast<rule*>(handle);
    auto it = _rules.begin();
    while (&*it != r && it != _rules.end())
        ++it;
    if (it == _rules.end())
        return;

    /* remove the rule from the loggers */
    _default.remove_rule(*it);
    for (auto& l : _logger)
        l.second->remove_rule(*it);

    /* remove the rule from the manager */
    _rules.erase(it);
}

void manager::reset()
{
    std::lock_guard<std::mutex> lg(_mutex);
    _logger.clear();
    _rules.clear();
    _consumer.clear();
}

#ifdef CPPLOGGING_HAS_LOAD_CONFIG
matcher_ptr_u load_matcher(const ::nlohmann::json& jMatchers, const ::nlohmann::json& jName)
{
    if (!jName.is_string())
        throw std::invalid_argument("expected matcher value to be an string");

    auto& jMatcher = jMatchers[jName.get<std::string>()];
    if (!jMatcher.is_object())
        throw std::invalid_argument("expected matcher to be an object");

    auto& jType = jMatcher["type"];
    if (!jType.is_string())
        throw std::invalid_argument("expected 'matcher.type' to be an string");

    /* match all cosumers/loggers */
    auto type = jType.get<std::string>();
    if (type == "all")
    {
        return std::make_unique<matcher_all>();
    }

    /* match the default logger */
    else if (type == "default_logger")
    {
        return std::make_unique<matcher_default_logger>();
    }

    /* regex */
    else if (type == "regex")
    {
        /* read regex value */
        auto& jValue = jMatcher["value"];
        if (!jValue.is_string())
            throw std::invalid_argument("expected 'matcher.value' to be an string");
        auto regex = jValue.get<std::string>();

        /* read regex invert */
        bool invert = false;
        auto& jInvert = jMatcher["invert"];
        if (!jValue.empty() && !jValue.is_boolean())
            throw std::invalid_argument("expected 'matcher.invert' to be an bool");
        invert = jInvert.get<bool>();

        return std::make_unique<matcher_regex>(regex, invert);
    }

    /* invalid type */
    else
    {
        using namespace ::std;
        throw std::invalid_argument("unknown matcher type: "s + type);
    }
}

log_level load_log_level(const ::nlohmann::json& jLogLevel, log_level default_level)
{
    if (jLogLevel.empty())
        return default_level;
    if (!jLogLevel.is_string())
        throw std::invalid_argument("expected rule log level to be an string");

    auto tmp = jLogLevel.get<std::string>();
    if (tmp == "debug")
        return log_level::debug;
    else if (tmp == "info")
        return log_level::info;
    else if (tmp == "warn")
        return log_level::warn;
    else if (tmp == "error")
        return log_level::error;
    else
    {
        using namespace ::std;
        throw std::invalid_argument("invalid log level: "s + tmp);
    }
}

void manager::load(const std::string& filename)
{
    using namespace ::nlohmann;

    /* load the json object */
    std::ifstream ifs(filename);
    json root;
    ifs >> root;

    /* get root nodes */
    auto rules = root["rules"];
    auto matchers = root["matchers"];
    auto consumers = root["consumers"];

    /* check root nodes */
    if (rules.empty() || !rules.is_array())
        throw std::invalid_argument("expected 'rules' to be an json array");
    if (matchers.empty() || !matchers.is_object())
        throw std::invalid_argument("expected 'matchers' to be an json object");
    if (consumers.empty() || !consumers.is_object())
        throw std::invalid_argument("expected 'consumers' to be an json object");

    /* read consumers */
    for (auto& c : consumers.items())
    {
        auto name = c.key();
        auto obj  = c.value();
        if (!obj.is_object())
            throw std::invalid_argument("expected 'consumer' to be an json object");

        /* type */
        auto& jType = obj["type"];
        if (!jType.is_string())
            throw std::invalid_argument("expected 'consumer.type' to be an string");

        /* format */
        auto& jFormat = obj["format"];
        auto format = consumer::default_format();
        if (jFormat.empty())
        {
            if (!jFormat.is_string())
                throw std::invalid_argument("expected 'consumer.format' to be an string");
            format = jFormat.get<std::string>();
        }

        /* log to file */
        auto type = jType.get<std::string>();
        if (type == "file")
        {
            auto& jFilename = obj["filename"];
            if (!jFilename.is_string())
                throw std::invalid_argument("expected 'consumer.filename' to be an string");

            auto fn = jFilename.get<std::string>();
            register_consumer(std::make_unique<consumer_stream>(
                name,
                std::make_unique<std::ofstream>(fn),
                format));
        }

        /* log to stdout */
        else if (type == "stdout")
        {
            register_consumer(std::make_unique<consumer_stream>(
                name,
                std::cout,
                format));
        }

        /* log to stderr */
        else if (type == "stderr")
        {
            register_consumer(std::make_unique<consumer_stream>(
                name,
                std::cerr,
                format));
        }

        /* invalid consumer type */
        else
        {
            using namespace ::std;
            throw std::invalid_argument("unknown consumer type: "s + type);
        }
    }

    /* read rules */
    for (auto& r : rules)
    {
        define_rule(
            load_matcher(matchers, r["logger"]),
            load_matcher(matchers, r["consumer"]),
            load_log_level(r["min"], log_level::debug),
            load_log_level(r["max"], log_level::error));
    }
}
#endif

logger_impl& manager::init_logger(logger_impl& logger)
{
    for (auto& rule : _rules)
    {
        if (rule.logger_matcher->match(logger))
            logger.add_rule(rule);
    }
    return logger;
}
